"""Related to implementing WET"""
import json

from django import template
from django import urls
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.template import loader
from django.templatetags.static import static
from django.urls.exceptions import NoReverseMatch
from django.utils.html import format_html

register = template.Library()

WET_CDN_ROOT = "https://cdn.jsdelivr.net/gh/wet-boew"


def jsdelivr(pkg, asset):
    """Construct a jsdelivr CDN URL for the provided WET package and asset."""
    ver = settings.WET_VERSION if pkg == "wet-boew" else settings.THEME_VERSION
    return f"{WET_CDN_ROOT}/{pkg}-dist@{ver}/{asset}"


@register.simple_tag
def phac_aspc_wet_css(base_only=False):
    """Generate the CSS tags required for WET

    If base_only is True, only those classes required for library features
    will be included.  (For example displaying the session timeout dialog).

    This should be used in the HEAD section of your templates.
    """
    css_url = (
        static("phac_aspc_helpers/base.css")
        if base_only
        else jsdelivr("themes", "GCWeb/css/theme.min.css")
    )
    no_script = jsdelivr("wet-boew", "wet-boew/css/noscript.min.css")
    return format_html(
        (
            f'<link rel="stylesheet" href="{css_url}">'
            f'<noscript><link rel="stylesheet" href="{no_script}"></noscript>'
        )
    )


@register.simple_tag
def phac_aspc_wet_scripts(include_jquery=True):
    """Generate the script tags required for WET

    If include_jquery is False, jquery will not be included

    This should be used directly before the closing </body> tag in your
    templates.
    """
    jquery = (
        """
      <script
        src="https://ajax.googleapis.com/ajax/libs/jquery/2.2.4/jquery.min.js"
        integrity="sha384-rY/jv8mMhqDabXSo+UCggqKtdmBfd3qC2/KvyTDNQ6PcUJXaxK1tMepoQda4g5vB"
        crossorigin="anonymous"
      ></script>"""
        if include_jquery
        else ""
    )
    wet_js = jsdelivr("wet-boew", "wet-boew/js/wet-boew.min.js")
    gcweb_js = jsdelivr("themes", "GCWeb/js/theme.min.js")
    return format_html(
        f"""
        {jquery}
        <script src="{ wet_js }"></script>
        <script src="{ gcweb_js }"></script>
    """
    )


@register.simple_tag(takes_context=True)
def phac_aspc_wet_session_timeout_dialog(context, logout_url):
    """Displays a dialog to the user warning them their session is about
    to expire, with the option to continue or end their session

    WARNING: Wet expects your page to have at least 1 H1 element, if not
    this component will not behave properly.  For this reason if no h1 is
    present one is automatically appended to the document with its display set
    to none.
    """
    if not context["request"].user.is_authenticated:
        return ""

    session_alive = settings.SESSION_COOKIE_AGE * 1000
    reaction_time = 180000 if session_alive >= 300000 else session_alive * 0.2

    logouturl = logout_url
    try:
        logouturl = urls.reverse(logout_url)
    except NoReverseMatch:
        pass

    try:
        return loader.get_template("phac_aspc/helpers/wet/session_timeout.html").render(
            {
                "config": json.dumps(
                    {
                        "inactivity": session_alive - reaction_time,
                        "reactionTime": reaction_time,
                        "refreshCallbackUrl": urls.reverse("phac_aspc_helpers_session"),
                        "method": "PUT",
                        "sessionalive": session_alive,
                        "logouturl": logouturl,
                    }
                )
            },
            request=context["request"],
        )
    except NoReverseMatch as exc:
        raise ImproperlyConfigured("The WET urls are not loaded.  See README") from exc



"""Provides all URLs in a single module"""
from .wet import urlpatterns as wet_patterns
from .auth import urlpatterns as auth_patterns

urlpatterns = wet_patterns + auth_patterns

__all__ = ['urlpatterns',]



"""Views related to WET"""
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse, HttpResponseNotFound


@csrf_exempt
def session(request):
    """Session refresh handler
    If the request is authenticated then the session is extended and the text
    "true" is returned, otherwise "false" is returned.
    """
    if request.method == "PUT":
        if request.user.is_authenticated:
            request.session.modified = True
            return HttpResponse("true")

        return HttpResponse("false")

    return HttpResponseNotFound()



"""This file contains utilities helpful for settings declarations"""
import os
import inspect

import environ

from django.core import checks
from django.conf import settings


def trigger_configuration_warning(msg, **kwargs):
    """Trigger a configuration error programmatically"""

    def phac_aspc_conf_warning(
        app_configs, **conf_kwargs
    ):  # pylint: disable=unused-argument
        return [checks.Warning(msg, **kwargs)]

    checks.register(phac_aspc_conf_warning)


def warn_and_remove(items, dest_list):
    """If the items intersects with dest_list, raise a configuration warning and
    remove it from the list returned to prevent errors."""
    ret = []
    context = inspect.stack()[1].function
    hint = "You should remove it from your project to ensure proper ordering."

    for item in items:
        if item in dest_list:
            trigger_configuration_warning(
                f"{item} is already defined in the provided list.",
                id=context,
                hint=hint,
            )
            hint = None
        else:
            ret.append(item)
    return ret


def configure_apps(app_list):
    """Return an application list which includes the apps required by this
    library"""

    prefix_list = warn_and_remove(["modeltranslation", "axes"], app_list)
    suffix_list = warn_and_remove(["phac_aspc.django.helpers"], app_list)

    return prefix_list + app_list + suffix_list


def configure_authentication_backends(backend_list):
    """Return the authentication backend list includes those required by this
    library.

    By default importing the settings will automatically configure the backend,
    however if you want to customize the authentication backend used by your
    project, you can use this method to ensure proper configuration."""
    oauth_backend = (
        [getattr(settings, "PHAC_ASPC_OAUTH_USE_BACKEND", "")]
        if getattr(settings, "PHAC_ASPC_OAUTH_USE_BACKEND", "")
        else []
    )

    prefix_backends = warn_and_remove(
        ["axes.backends.AxesStandaloneBackend"] + oauth_backend, backend_list
    )
    return prefix_backends + backend_list


def configure_middleware(middleware_list):
    """Return the list of middleware configured for this library"""
    prefix = warn_and_remove(
        ["axes.middleware.AxesMiddleware", "django.middleware.locale.LocaleMiddleware"],
        middleware_list,
    )
    return prefix + middleware_list


def get_env_value(env, key, prefix="PHAC_ASPC_"):
    """Return prefixed value from environment"""
    return env(f"{prefix}{key}")


def get_env(prefix="PHAC_ASPC_", **conf):
    """Return django-environ configured with the provided values and
    using the prefix.

    prefix can be used to change the environment variable prefix that is added
    to the beginning on the variables defined in conf.  By default this value is
    `PHAC_ASPC_`.

    conf is a dictionary used to generate the scheme for django-environ.

    See https://django-environ.readthedocs.io/en/latest/api.html#environ.Env for
    additional information on the scheme.
    """

    def _find_env_file(path):
        # look for .env file in provided path
        filename = os.path.join(path, ".env")
        if os.path.isfile(filename):
            return filename

        # search the parent
        parent = os.path.dirname(path)
        if parent and os.path.normcase(parent) != os.path.normcase(
            os.path.abspath(os.sep)
        ):
            return _find_env_file(parent)

        # Not found
        return ""

    scheme = {}
    for name, values in conf.items():
        scheme[f"{prefix}{name}"] = values

    env = environ.Env(**scheme)
    environ.Env.read_env(_find_env_file(os.path.abspath(os.path.dirname(__name__))))
    return env


def global_from_env(prefix="PHAC_ASPC_", **conf):
    """Create named global variables based on the provided environment variable
    scheme.  Variables defined in the scheme will be inserted into the calling
    module's globals and prefixed with `PHAC_ASPC_` when fetching the
    environment variable.

    prefix can be used to change the environment variable prefix that is added
    to the beginning on the variables defined in conf.  By default this value is
    `PHAC_ASPC_`.

    conf is a dictionary used to generate the scheme for django-environ.


    """

    mod = inspect.getmodule(inspect.stack()[1][0])
    env = get_env(prefix, **conf)

    for name in conf:
        setattr(mod, name, get_env_value(env, name))


def configure_settings_for_tests():
    """
    Modify django settings so they are suitable to run unit tests.
      - Disables the Axes authentcation backend, to allow the test client
        to function.  You could instead modify the request as described here:
        https://django-axes.readthedocs.io/en/latest/3_usage.html#authenticating-users
    """
    # settings.AXES_ENABLED = False
    settings.AUTHENTICATION_BACKENDS = [
        x
        for x in settings.AUTHENTICATION_BACKENDS
        if x != "axes.backends.AxesStandaloneBackend"
    ]



"""Locale switching is global to the application, so we do it in a thread-safe
context so no other parts of the application are affected."""

import locale
import threading
from contextlib import contextmanager

from django.utils.translation import activate, get_language

from .code import get_language_code

LOCALE_LOCK = threading.Lock()


@contextmanager
def locale_lang(lang):
    """Isolate changing of a locale to the code block that requires it.

    Usage:

    with locale_lang('fr-ca'):
        <insert your code here>

    """
    with LOCALE_LOCK:
        saved = locale.setlocale(locale.LC_ALL)
        old_lang = get_language()
        try:
            code = get_language_code(lang)
            activate(lang)
            yield locale.setlocale(
                locale.LC_ALL, "fr_CA.UTF8" if code == "fr" else "en_CA.UTF8"
            )
        finally:
            locale.setlocale(locale.LC_ALL, saved)
            activate(old_lang)



"""
Ready unit tests
"""
from unittest.mock import MagicMock

from phac_aspc.django.helpers.ready import (
    process_ready_hooks,
    execute_when_ready,
    ready_hooks,
)


def test_execute_ready():
    """
    Test that functions registered with `execute_when_ready` are added to the
    queue, called by `process_ready_hooks` and removed.
    """
    process_ready_hooks()  # Make sure there are no lingering hooks
    func = MagicMock()
    execute_when_ready(func)
    assert len(ready_hooks) == 1
    assert ready_hooks[0] == func
    process_ready_hooks()
    assert func.called
    assert len(ready_hooks) == 0



"""Hooks module"""
__author__ = "Luc Belliveau <luc.belliveau@phac.aspc.gc.ca>"
model_translation_hooks = []


def process_translate_decorators():
    """
    Process the registered hooks.

    This should occur after the models are loaded by Django, for example from
    the AppConfig's ready function.
    """
    for hook in model_translation_hooks:
        hook()

    model_translation_hooks.clear()



class PytestRunner:
    """Runs pytest to discover and run tests."""

    @classmethod
    def add_arguments(cls, parser):
        parser.set_defaults(keepdb=True)

        # remaps to the -k cli arg of pytest, useful for test selection
        parser.add_argument(
            "-s",
            "--select",
            help="remaps to -k test-selection argument in pytest",
        )

    def __init__(
        self,
        verbosity=1,
        failfast=False,
        keepdb=True,
        select=None,
        **kwargs,
    ):
        self.verbosity = verbosity
        self.failfast = failfast
        self.keepdb = keepdb
        self.select = select

    def run_tests(self, test_labels):
        """Run pytest and return the exitcode.

        It translates some of Django's test command option to pytest's.
        """
        import pytest

        argv = []
        if self.select is not None:
            argv.append(f"-k {self.select}")
        if self.verbosity == 0:
            argv.append("--quiet")
        if self.verbosity == 2:
            argv.append("--verbose")
        if self.verbosity == 3:
            argv.append("-vv")
        if self.failfast:
            argv.append("--exitfirst")
        if self.keepdb:
            argv.append("--reuse-db")

        argv.extend(test_labels)
        return pytest.main(argv)



# pylint: disable=missing-function-docstring, missing-class-docstring
"""
utilities for writing excel data using openpyxl 
"""
import re

from django.core.paginator import Paginator
from django.views import View
from django.http import HttpResponse
from django.utils.functional import Promise
from django.utils.safestring import SafeString

try:
    import openpyxl
    from openpyxl.cell.cell import ILLEGAL_CHARACTERS_RE
    from openpyxl.utils import escape as escapeSvc
except (ImportError, ModuleNotFoundError) as exc:
    raise ImportError("must install openpyxl==3.0.10 to use excel helpers") from exc

# Note that OPENPXYL starts columns and rows at index 1.


GLOBAL_MAX_COL_LEN = 5000


def escape_for_xlsx(text):
    """
    escape a string for excel
    """
    # pylint: disable=unidiomatic-typecheck
    return escapeSvc.escape(text) if type(text) == str else text


def create_field_val_getter(field_name):
    """
    returns a function that gets the value of a field for a django model instance
    """
    return lambda record: record.serializable_value(field_name)


def write_queryset_to_sheet(workbook, queryset):
    """
    iterates over queryset and puts all fields the workbook in a sheet
    - picks scalars and FK-IDs based on model
    - picks sheet name based on model
    - warning: uses queryset.iterator, prefetches are ignored
    """

    ModelToSheetWriter(workbook=workbook, queryset=queryset).write()


def page_queryset(queryset, per_page=1000):
    """
    While in a perfect world you would use queryset.iterator()
    but there is an issue that it fails to
    load any prefetch_related() fields specified.
    Paginator() can mimic the same functionality

    https://github.com/django-import-export/django-import-export/issues/774#issuecomment-449064652
    """

    # pylint: disable=protected-access
    if queryset._prefetch_related_lookups:
        if not queryset.query.order_by:
            # Paginator() throws a warning if there is no sorting attached to the queryset
            queryset = queryset.order_by("pk")
        paginator = Paginator(queryset, per_page)
        for index in range(paginator.num_pages):
            yield from paginator.get_page(index + 1)
    else:
        yield from queryset.iterator(chunk_size=per_page)


class Column:
    """
    Base class to write columns in a sheet
    """

    def get_header(self):
        """return a header string for the column"""
        raise NotImplementedError()

    def get_value(self, record):
        """return this column's intended cell value for a record"""
        raise NotImplementedError()

    def serialize_value(self, value):
        return serialize_value(value)

    def get_serialized_value(self, record):
        return self.serialize_value(self.get_value(record))


class ModelColumn(Column):
    """
    shorthand for defining a column that writes a scalar model field (non foreign-key)
    """

    def __init__(self, model_cls, field_name, header_value=None):
        self.header_value = header_value
        self.model_cls = model_cls
        self.field_name = field_name
        self.get_val = create_field_val_getter(field_name)

    def get_header(self):
        # pylint: disable=protected-access
        header_value = (
            self.header_value
            or self.model_cls._meta.get_field(self.field_name).verbose_name
        )
        return escape_for_xlsx(header_value)

    def get_value(self, record):
        return self.get_val(record)


class ChoiceColumn(Column):
    def __init__(self, model_cls, field_name, header_value=None):
        self.header_value = header_value
        self.model_cls = model_cls
        self.field_name = field_name

    def get_header(self):
        # pylint: disable=protected-access
        header_value = (
            self.header_value
            or self.model_cls._meta.get_field(self.field_name).verbose_name
        )
        return escape_for_xlsx(header_value)

    def get_value(self, record):
        method_name = "get_" + self.field_name + "_display"
        return getattr(record, method_name)()


class CustomColumn(Column):
    def __init__(self, header, get_val):
        self.header = header
        self.get_val = get_val

    def get_header(self):
        return self.header

    def get_value(self, record):
        return self.get_val(record)


class ManyToManyColumn(Column):
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        model: type,
        field_name: str,
        get_related_str: callable = None,
        delimiter: str = ", ",
        header=None,
    ):
        self.model = model
        self.field_name = field_name
        self.header = header
        if get_related_str:
            self.get_related_str = get_related_str
        else:
            self.get_related_str = lambda x, *args: str(x)

        self.delimiter = delimiter

    def get_header(self):
        if self.header:
            return self.header
        # pylint: disable=protected-access
        return self.model._meta.get_field(self.field_name).verbose_name

    def get_value(self, record):
        related_records = list(getattr(record, self.field_name).all())
        return self.delimiter.join([self.get_related_str(x) for x in related_records])


class ModelToSheetWriter:
    def __init__(self, workbook, queryset):
        self.workbook = workbook
        self.queryset = queryset
        self.model = queryset.model

    def get_column_configs(self):
        # pylint: disable=protected-access
        fields_to_write = self.model._meta.fields
        return [ModelColumn(self.model, field.name) for field in fields_to_write]

    def get_sheet_name(self):
        return get_default_sheet_name_for_qs(self.queryset)

    def get_header_row(self):
        # return [escape_for_xlsx(name) for name, _ in self.get_column_configs()]
        return [serialize_value(col.get_header()) for col in self.get_column_configs()]

    def write(self):
        worksheet = self.workbook.create_sheet(title=self.get_sheet_name())

        worksheet.append(self.get_header_row())

        iterator = page_queryset(self.queryset)
        for record in iterator:
            xl_row = []
            for col in self.get_column_configs():
                xl_val = col.get_serialized_value(record)
                xl_row.append(xl_val)

            worksheet.append(xl_row)


def serialize_value(value):
    if value is None:
        write_val = ""
    else:
        if value is True:
            write_val = 1
        elif value is False:
            write_val = 0
        elif isinstance(value, list):
            write_val = str(value)
        elif isinstance(value, (Promise, SafeString)):
            # tm values don't play well with excel
            write_val = value + ""
        else:
            write_val = value

    xl_val = escape_for_xlsx(write_val)
    if next(ILLEGAL_CHARACTERS_RE.finditer(str(value)), None):
        xl_val = re.sub(ILLEGAL_CHARACTERS_RE, "", str(xl_val))

    return xl_val


def get_manager_for_export(model):
    return getattr(model, "no_fakes", model.objects)


def get_default_sheet_name_for_qs(queryset):
    """
    excel only supports up to 31 characters in a worksheet name
    """
    # pylint: disable=protected-access
    model_name = queryset.model._meta.verbose_name
    return f"{model_name[:30]}"


class AbstractExportView(View):
    def get_filename(self):
        return "export.xlsx"

    def get_queryset(self):
        try:
            return self.queryset
        except AttributeError as e:
            raise NotImplementedError(
                "Must define queryset attr or override get_queryset()"
            ) from e

    def get_sheetwriter_class(self):
        try:
            return self.sheetwriter_class
        except AttributeError as e:
            raise NotImplementedError(
                "Must define sheetwriter_class attr or override get_sheetwriter_class()"
            ) from e

    def get(self, request, *args, **kwargs):
        wb = openpyxl.Workbook(write_only=True)
        WriterCls = self.get_sheetwriter_class()
        writer = WriterCls(
            workbook=wb,
            queryset=self.get_queryset(),
        )
        writer.write()
        response = HttpResponse(headers={"Content-Type": "application/vnd.ms-excel"})
        response["Content-Disposition"] = f"attachment; filename={self.get_filename()}"
        wb.save(response)
        return response



"""OAuth authentication related views"""
from urllib import parse

from django.conf import settings
from django.contrib.auth import authenticate
from django.contrib.auth import login as auth_login
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse

from authlib.integrations.django_client import OAuth, OAuthError

oauth = OAuth()

PROVIDER = getattr(settings, "PHAC_ASPC_HELPER_OAUTH_PROVIDER", False)
BACKEND = getattr(settings, "PHAC_ASPC_OAUTH_USE_BACKEND", False)
REDIRECT_LOGIN = getattr(settings, "PHAC_ASPC_OAUTH_REDIRECT_ON_LOGIN", "")

if PROVIDER:
    oauth.register(PROVIDER)


def validate_iss(claims, value):
    """Validate the iss claim"""
    tenant = getattr(settings, "OAUTH_MICROSOFT_TENANT", "common")
    use_tenant = tenant if tenant != "common" else claims["tid"]
    # iss = "https://login.microsoftonline.com/{}/v2.0".format(use_tenant)
    return use_tenant in value


def login(request):
    """Redirect users to the provider's login page"""
    if PROVIDER:
        client = oauth.create_client(PROVIDER)
        auth_url_extra_params = {"state": request.build_absolute_uri()}
        return client.authorize_redirect(
            request,
            request.build_absolute_uri(reverse("phac_aspc_authorize")),
            **auth_url_extra_params
        )
    raise ImproperlyConfigured("The login route is not configured.")


def authorize(request):
    """Verify the token received and perform authentication"""
    if PROVIDER:
        try:
            client = oauth.create_client(PROVIDER)
            token = client.authorize_access_token(
                request,
                claims_options={"iss": {"essential": True, "validate": validate_iss}},
            )
            user_info = token["userinfo"]
            query_params = dict(
                parse.parse_qsl(parse.urlsplit(request.GET["state"]).query)
            )
            user = authenticate(request, user_info=user_info)
            if user is not None:
                auth_login(
                    request,
                    user=user,
                    backend=BACKEND,
                )

                if "next" in query_params:
                    return HttpResponseRedirect(query_params["next"])

                return HttpResponseRedirect(
                    reverse(REDIRECT_LOGIN) if REDIRECT_LOGIN else "/"
                )
            return render(
                request,
                "phac_aspc/helpers/auth/error.html",
                {
                    "type": "oauth",
                    "details": "Access denied",
                },
            )

        except OAuthError as err:
            return render(
                request,
                "phac_aspc/helpers/auth/error.html",
                {
                    "type": "oauth",
                    "details": err.description,
                },
            )
        except Exception as err:  # pylint: disable=broad-except
            return render(
                request,
                "phac_aspc/helpers/auth/error.html",
                {
                    "type": "general",
                    "details": str(err),
                },
            )
    raise ImproperlyConfigured("The authorize route is not configured.")



"""Related to implementing WET"""
from django import template
from django.utils.translation import get_language

register = template.Library()


@register.simple_tag()
def phac_aspc_localization_lang():
    """Returns the current language code"""
    return get_language()


@register.simple_tag(takes_context=True)
def use_string(context, name, strings=None):
    """
    Loads the string from a template or via the variable dict `strings` if the
    `name` key is defined within.  This allows strings to be overridden in 2
    ways, either by user defined templates or via the render context.

    When `name` is not found in `strings`, the template name becomes:

    phac_aspc/helpers/strings/{name}.html

    """
    if strings and name in strings:
        return strings[name]

    return template.loader.get_template(
        f"phac_aspc/helpers/strings/{name}.html", using="django"
    ).render(context.flatten())



"""Related to Authentication"""
from django import template
from django.conf import settings
from django.template import loader

register = template.Library()


@register.simple_tag()
def phac_aspc_auth_signin_microsoft_button(query_params=""):
    """Returns a signin button using the microsoft design"""

    return (
        loader.get_template("phac_aspc/helpers/auth/buttons/microsoft.html").render(
            {"query_params": query_params}
        )
        if getattr(settings, "PHAC_ASPC_HELPER_OAUTH_PROVIDER", False)
        else ""
    )



import factory

from .models import Book, Author, Tag


class AuthorFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Author

    first_name = factory.Faker("bs")
    last_name = factory.Faker("bs")


class TagFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Tag

    name = factory.Faker("bs")


class BookFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Book

    author = factory.SubFactory(AuthorFactory)
    title = factory.Faker("bs")



# pylint: disable=protected-access
"""Test admin decorators"""
from django.db import models
from django.contrib.admin import site

from phac_aspc.django.helpers.ready import process_ready_hooks
from phac_aspc.django.admin.decorators.admin_decorators import add_admin


def test_add_admin_decorator():
    """Ensure add_admin decorator works"""

    @add_admin()
    class TestModel(models.Model):
        """Example Model to be tested"""

        name = models.TextField()

    class TestOtherModel(models.Model):
        """Example Model to not be tested"""

        name = models.TextField()

    num_registered_before = len(site._registry)
    process_ready_hooks()

    assert len(site._registry) == num_registered_before + 1

    assert TestModel in site._registry

    assert TestOtherModel not in site._registry



"""Setup URLs for views related to authentication"""
from django.urls import path
from django.conf import settings
from django.conf.urls.i18n import i18n_patterns

from ...helpers.auth.views import login, authorize

urlpatterns = (
    i18n_patterns(
        path("phac_aspc_helper_login", login, name="phac_aspc_helper_login"),
        path("phac_aspc_helper_authorize", authorize, name="phac_aspc_authorize"),
    )
    if getattr(settings, "PHAC_ASPC_HELPER_OAUTH_PROVIDER", False)
    else []
)



from phac_aspc.django.excel import (
    ModelToSheetWriter,
    ModelColumn,
    CustomColumn,
    ManyToManyColumn,
    AbstractExportView,
)
from django.test import RequestFactory
import random

from openpyxl import Workbook
from testapp.models import Book
from testapp.model_factories import TagFactory, BookFactory


def create_data():
    all_tags = [TagFactory() for _ in range(6)]

    for _ in range(30):
        book = BookFactory()
        tags = random.sample(all_tags, random.randint(0, 4))
        book.tags.set(tags)


def test_model_to_sheet_writer(django_assert_max_num_queries):
    create_data()

    columns = [
        ModelColumn(Book, "title"),
        CustomColumn("Author", lambda x: f"{x.author.first_name} {x.author.last_name}"),
        ManyToManyColumn(Book, "tags"),
    ]

    class BookSheetWriter(ModelToSheetWriter):
        model = Book

        def get_column_configs(self):
            return columns

    with django_assert_max_num_queries(4):
        wb = Workbook()
        writer = BookSheetWriter(
            workbook=wb, queryset=Book.objects.all().prefetch_related("author", "tags")
        )
        writer.write()


def test_abstract_view():
    create_data()

    class BookSheetWriter(ModelToSheetWriter):
        model = Book

        def get_column_configs(self):
            return [
                ModelColumn(Book, "title"),
                CustomColumn(
                    "Author", lambda x: f"{x.author.first_name} {x.author.last_name}"
                ),
                ManyToManyColumn(Book, "tags"),
            ]

    class BookExportView(AbstractExportView):
        sheetwriter_class = BookSheetWriter
        queryset = Book.objects.all().prefetch_related("author", "tags")

    view_func = BookExportView.as_view()
    req_factory = RequestFactory()
    request = req_factory.get("/fake-url")

    response = view_func(request)
    assert response.status_code == 200
